// --- [ DCF_77.c ] -------------------------------------------------------------
//       25.10.09
//       tab = 3
//
// neu: das Loggen der Daten via Serielle Schnittstelle
//      das Loggen der zutreffenden Alarmdefinitionen
// ----------------------------------------------------------------------------
#define EXTERN

#include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "global.h"
#include "dcf_bits.h"
#include "dcf_77.h"
#include "lcd.h"
#include "sun.h"
#include "TWI_master_slave.h"
#include "rs232.h"

uint8_t summand[] 		PROGMEM = { 1, 2, 4, 8, 10, 20, 40, 80 }; // fr dcf77
uint8_t days_of_month[] PROGMEM = { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
char day_names[] 			PROGMEM = {'M','o','D','i','M','i','D','o','F','r','S','a','S','o' };

volatile uint8_t flag;						// sammelt Status-Flags
volatile uint8_t error;						// sammelt error-Flags
volatile uint8_t T0_ticks;					// Counter fr die interne Uhr / Sekundensignal
volatile uint8_t T1_ticks;					// Counter fr die Lnge des DCF-Signals (in 0.01sek)
uint8_t error_abs;							// summiert die Anzahl der fehlerhaften Minuten
													// wird um 0.00h zurckgesetzt

// ----------------------------------------------------------------------------
int main(void)
{
// --- INITIALISIERUNG --------------------------------------------------------
uint8_t mcusr;									// Flags in MCUSR nach dem Start
//uint8_t start_flag = TRUE;					// damit ein Neustart erkannt werden kann

// --- PORTs ------------------------------------------------------------------
// PD.0/1	RxD, TxD 						// fr serielle Schnittstelle 
// PD.2/3	INT0, INT1						// hier wird die Uhr angeschlossen
// PD.4     frei (1)
// PB.6/7	XTAL								// der Quarz bentigt zwei Pin
// PD.5..7	frei (3)							// PD.7 als Schalter genutzt -> DAYLIGHT_PIN
// PB.0..2  LCD 								// Steuerleitungen LCD siehe lcd.h
// PB.4/5   frei (2)
// PC.0..3  LCD								// Datenleitungen LCD  siehe lcd.h
// PC.4/5	SCL,SDA							// TWI-Port

// erwartet wird vom DCF-Empfnger:
// DCF-Ruhe-Level: HIGH, bei Signal: LOW !!!!!!!!!!!!!!!!!!!!!!!!!

uint8_t i;								

mcusr = MCUSR;									// MCUSR sichern
MCUSR = 0x0F;									// alle Flags in MCUSR lschen

PORTD = 1<<DAYLIGHT_PIN;
DDRD  = 1<<DAYLIGHT_PIN;					// nur PIN.7 als Alarmausgang benutzt
													// PD.0/PD.1 = RxD/TxD
//DDRD = (1<<PD0| 1<<PD1 | 1<<PD4 | 1<<PD5| 1<<PD6| 1<<PD7);

// --- ADC --------------------------------------------------------------------
PRR = (1<<PRSPI | 0<<PRUSART0 | 1<<PRADC);// SPI, (USART) und ADC abschalten

// --- EXT_INT ----------------------------------------------------------------
#ifdef M168
EICRA = (1<<ISC11 | 1<< ISC10 | 1<< ISC01);// Rising INT1, Falling INT0
EIMSK  = (1<<INT1 | 1<< INT0);			// INT0 / INT1_isr enablen
#endif
#ifdef M8
MCUCR = (1<<ISC11 | 1<< ISC10 | 1<< ISC01);// Rising INT1, Falling INT0
GICR  = (1<<INT1 | 1<< INT0);			// INT0 / INT1_isr enablen
#endif

// --- CLock-Prescaler --------------------------------------------------------
#ifdef M168
CLKPR = (1<<CLKPCE);
CLKPR &= 0xF0;									// lower nibble lschen -> Faktor 1
#endif

// --- TIMER0 -----------------------------------------------------------------
// generiert den internen 1-Sekunden-Takt
// 3.686.400 / ((64 * 256) * 225) = 1 Sekunde
TCCR0B = (1<<CS01 | 1<< CS00);			// Prescale 64, den Rest macht TIMER0_ISR
#define T0_PRELOAD	225
// --- TIMER1 -----------------------------------------------------------------
// luft als Zhler fr Pulsdauer und Pulsabstand
// 3.686.400 / (256 * 144)`= 0.01 Sekunde
TCCR1A = 0;
TCCR1B = (1<<CS12);							// Prescale 256
OCR1AH = 0;
OCR1AL = 144;

// --- beide Timer ------------------------------------------------------------
#ifdef M8
//TIMSK = (1<<TOIE0 | 1<<OCIE1A);		// ISR fr Timer1 / 2 enablen
#endif
#ifdef M168
TIMSK0 = (1<<TOIE0 );						// ISR fr Timer1 enablen
TIMSK1 = (1<<OCIE1A);						// ISR fr Timer2 enablen
#endif

asm volatile("sei");
TWI_Initialise();
DCF_read_alarm();								// gespeicherte Schaltdefinitionen aus dem Eeprom einlesen

#ifdef LOG_RS232
rs232_init(TEILER_9600);
put_c(13);
puts_p(PSTR("Logging on RS232"));		// "Hallo World" ausgeben
put_c(13);

// alle Alarmdefinitionen ausgeben
for (i = CLK_ARRAYS; i < (ALARMDEFS + CLK_ARRAYS); i++)
	{
	if (time[i][MODUS])
		{
		print_alarm(i);
		put_c(13);
		}
	}

#endif

// --- intro ------------------------------------------------------------------
// liefert im gnstigsten Fall nach 62 Sekunden ein vollstndiges Datum,
// im ungnstigsten Fall nach 120 Sekunden


#ifdef LCD
lcd_init(LCD_DISP_ON);
lcd_puts_p(PSTR("wait for synch .."));// Hinweis ausgeben

lcd_goto_xy(2,1);
lcd_puts_p(PSTR("____________________"));// Linie ausgeben

lcd_goto_xy(4,1);
lcd_puts_p(PSTR("TWI_ADR: "));			// aktuelle TWI-Adresse ausgeben
lcd_dezprint(TWAR);

#ifdef WATCHDOG
if (mcusr & (1<<WDRF))						// wenn sich ein watchdog-reset ereignet hat
	{
	MCUSR = 0x0F;								// Flag lschen
	lcd_goto_xy(3,1);
	lcd_puts("wd");							// "wd" in Zeile 3 ausgeben
	}
#endif

#endif

// --- warten auf den Minutenanfang -------------------------------------------
//
while (!(flag & (1<<DCF_MINUTE)))		// solange warten, bis das Minutenflag gesetzt ist
	{									
	if (flag & (1<<CLK_SEC))				// die Sekunden seit dem Start anzeigen
		{
		flag &= (~(1<<CLK_SEC));
		if (error) error = 0;				// Fehler sollen die Synchronisation mit dem 
													// Minutenbeginn nicht stren
													// Timer0 setzt bei Fehler kein Minutenflag
		#ifdef LCD
		display_second();						// nur die Sekundenanzeige aktualisieren
													// Damit der Zhler weiterluft, wenn bis hierhin
													// kein Minutensignal gefunden wurde.
		if (time[CLK][SEKUNDE] == 58) time[CLK][SEKUNDE] = 0;
		#endif
		}
	}

// --- Minutenanfang erkannt --------------------------------------------------
#ifdef ZEIT_ZEICHEN
time[CLK][STUNDE] = 16;						// damit das Zeitzeichen fr 4:00 ausgegeben wird
zeit_zeichen();								// gibt das Zeitzeichen aus
#endif
													// jetzt ist das Minutenflag der DCF-Uhr gesetzt
time[CLK][STUNDE]  = 0;
time[CLK][SEKUNDE] = 0;						// wieder bei Sekunde 0 starten,
DCF_reset_bits();									// den Index von DCF_bits auf 0 setzen
error 	= 0;
flag 		= (1<<START_FLAG);
T1_ticks = 0;									// alle Counter zurcksetzen
TCNT1H   = 0;
TCNT1L   = 0;

#ifdef LCD
lcd_goto_xy(1,1);							
lcd_puts_p(PSTR("new minute found:"));	// Hinweis ausgeben
lcd_goto_xy(4,1);
lcd_puts_p(PSTR("            "));		// Text 'TWI-Adr' lschen	
#endif


// --- Watchdog ---------------------------------------------------------------
// den Watchdog erst dann enablen, wenn ein DCF-Signal gefunden ist
// sonst wird bei fehlendem Signal nach 59 Sekunden ein reset ausgefhrt.
// der Watchdog wird zurckgesetzt in main bei flag &= (1<<CLK_SEC)
#ifdef WATCHDOG
asm volatile("cli");
WDTCSR = (1<<WDCE | 1<<WDE);
WDTCSR = (0<<WDCE | 1<<WDE | 1<<WDP3);	// 4 Sekunden (mindestens 2 Sekunden sind erforderlich
asm volatile("sei");							// um die Zeit von der 59. Sekunde bis zur 1.Sekunde zu berbrcken)
#endif
													
// --- main -------------------------------------------------------------------
// eine Endlosschleife, in der die Flag.bits den Ablauf steuern

while(1)
	{

// --- neue Sekunde der xtal-Uhr ----------------------------------------------

	if (flag & (1<<CLK_SEC))
		{
		flag &= (~(1<<CLK_SEC));
		#ifdef LCD
		display_second();						// nur die Sekundenanzeige aktualisieren
		#endif
		#ifdef WATCHDOG
		asm volatile("wdr");
		#endif
		}

// --- DCF-Signal empfangen ---------------------------------------------------

													// wird direkt von ISR(INT1_vect) behandelt 

// --- 59. Sekunde erreicht, Vergleich der neuen Minute mit der aktuellen anstellen ---

	if (flag & (1<<CLK_SEC_59))			// nach der 59. Sekunde kann decodiert werden
		{
		flag &= ~(1<<CLK_SEC_59);			// flag lschen
		DCF_decode();							// die empfangengen DCF-Bits decodieren
													// nun steht die decodierte Zeit der neuen Minute in time[DCF]
													// in time[CLK] steht noch die aktuelle Minute
													// error kann noch durch Parittsfehler von DCF_decode() gesetzt werden
													// oder durch fehlerhafte Impulslngen whrend der letzten Minute
#ifdef LOG_RS232		
		print_datum(); 						// die Zeitdaten der aktuellen Minute ausgeben
		print_dcfbits();						// die DCF-Bits der folgenden Minute ausgeben
#endif													
													
		if (error)								// wenn ein Fehler aufgetreten ist, dann
			{										// die aktuelle Zeit um 1 Minute incrementieren
			DCF_incr_time(&time[CLK][0], 1);
			}
			
		else										// wenn kein Fehler im DCF-Signal erkannt worden ist, aber	
													// damit ist aber noch nicht die Abwesenheit eines Fehlers garantiert
													// daher wird die neue Zeit mit der letzten Minute verglichen 
			{									
													// wenn time[CMP] = (time[CLK] + 1) == time[DCF]
													// wenn letzte Minute + 1 == neue DCF-Zeit, dann alles in Ordnung
			DCF_copy_time(&time[CMP][0], &time[CLK][0]); // Uhr-Zeit kopieren
			DCF_incr_time(&time[CMP][0], 1); // und um eine Minute incrementieren
			if (DCF_compare_time(&time[CMP][0], &time[DCF][0])) // und mit der neuen DCF-Zeit vergleichen
				{									// wenn Vergleich erfolgreich, dann time[CMP] in die Uhr kopieren
				DCF_copy_time(&time[CLK][0], &time[CMP][0]);
				}
			
			else									// wenn nicht, dann geht's hier weiter
				{									// Vergleich der neuen mit incrementierter alten Minute in CMP_time.
													// hier gehts weiter, wenn ein Fehler beim Vergleich aufgetaucht ist
													// das Problem taucht IMMER beim Start in der 1. Minute auf
													// weil hier noch kein Vergleichswert einer Vorminute vorliegt
					
				if (flag & (1<<START_FLAG))// das Startflag ist nach einem Neustart gesetzt - und nur dann !!
					{
					flag &= ~(1<<START_FLAG);
					error = 0;					// Fehlerflag lschen, es wird keine Fehlermeldung angezeigt

					DCF_copy_time(&time[CLK][0], &time[DCF][0]);// dann die DCF-Werte in die Uhr kopieren
					}								//

				else								// wenn der Fehler nicht beim Programmstart auftritt
					{
					// als anderes Kriterium fr die Korrektheit des Zeitsignals:
					// zwei aufeinander folgende DCF-Signale unterscheiden sich um 1 Minute
					DCF_incr_time(&time[BAK][0], 1); // letzte DCF-Zeit (time[BAK][]) um eine Minute erhhen
													// wenn der Vergleich misslingt, dann Fehlercode setzen

					if (DCF_compare_time(&time[BAK][0], &time[DCF][0]))
						{							// wenn zwei aufeinander folgende DCF-Sequenzen zueinander passen 
													// dann dies als gltige neue Zeit bernehmen
						DCF_copy_time(&time[CLK][0], &time[DCF][0]);// die DCF-Werte der nchsten Minute in die Uhr kopieren
						error = ERR_SYNCH;	//
						}
						
					else							// Gelingt auch dieser Vergleich nicht, dann liegt ein Fehler vor.
						{							// Der tritt auch beim Wechsel von Sommer- auf
													// Winterzeit auf, wenn die Zeit um 1 Stunde "springt".
													// In diesem Falle muss das Ankndigungsbit A1 gesetzt sein
													// und die aktuelle MInute die 59. sein
													// genaugenommen auch noch Stunde == 3 (Sommer-nach Winter) bzw.
													// Stunde == 2 (Winter-nach-Sommer)
						if ((time[DCF][ZBITS] & 0x10) && (time[CLK][MINUTE] == 59))
							{
							error = 0;			// wenn A1-Bit gesetzt, dann wird keine Fehlermeldung generiert
													// auf weiteren Test wird hier verzichtet und 
							DCF_incr_time(&time[CLK][0], 1); // die Zeit manuell fortschreiben	
							}
							
						else						// wenn das A1-Bit nicht gesetzt ist, dann
							{						// liegt ein Fehler im Programm oder im DCF-Signal vor
							error = ERR_COMP_DCF;
							DCF_incr_time(&time[CLK][0], 1); // die Zeit manuell fortschreiben
							}
						}							
					}								
				}
			}
													// immer in der 59. Sekunde: 
		DCF_reset_bits();						// - die dcf_bits neu initialisieren
													//   dabei ein Kopie in DCF_bits[] zur Abholung ber TWI ablegen
													// - eine Kopie der DCF-Zeit in time[BAK] anlegen fr spteren Zeitvergleich
		DCF_copy_time(&time[BAK][0], &time[DCF][0]); 
		}

// --- Minuten-Anzeige aktualisieren ------------------------------------------

	else if (flag & (1<<CLK_MINUTE | 1<<DCF_MINUTE)) 	
		{
		flag &= ~(1<<CLK_MINUTE | 1<<DCF_MINUTE);	// Flags lschen
		time[CLK][SEKUNDE] = 0;				// Sekunden vor der Anzeige auf 0 stellen

		if (error)								// wenn in der letzten Minute ein Fehler aufgetreten sind,
			{										// dann kann hier reagiert werden
													// Fehlerflags erst nach der Anzeige lschen !!
			error_abs++;						// zum Debuggen die Fehler mitzhlen
			
#ifdef LOG_RS232
			put_s(" # ");
			put_s(itoa(error_abs, msg, 10));
			puts_p(PSTR(" err code: 0x"));
			put_s(itoa(error, msg, 16));
			if      (error == ERR_COMP_DCF)	puts_p(PSTR(", err: CLK-DCF compare"));
			else if (error == ERR_SYNCH   )  puts_p(PSTR(", err: resynch"));
			else if (error == ERR_NO_DCF  )  puts_p(PSTR(", err: timeout"));
			else if (error == ERR_TWI  )     puts_p(PSTR(", err: testtime"));
			else
				{
				if (error & 0xf8) puts_p(PSTR(", err: DCF-signal"));
				if (error & 0x07) puts_p(PSTR(", err: parity"));
				}	
#endif
			}

#ifdef LCD
		display_time(&time[CLK][0]);		// die komplette Anzeige neu aufbauen
#endif

#ifdef DAY_LIGHT
		DCF_check_daylight();				// Testet auf Sonnenauf-/Untergang und schaltet ggf. DAYLIGHT_PIN
#endif

#ifdef SUN_INCLUDE
		SUN_display();							// Fhrt die Kalkulation des Sonnenstandes aus
													// und zeigt die Daten an
#endif

#ifdef ZEIT_ZEICHEN
		zeit_zeichen();						// sendet bei vollen 15 Minuten ein Signal an einen TWI-Slave
													// ohne das dazu eine explizite Schaltdefinition vorliegen muss.
#endif
	
		DCF_check_alarm();					// Testen, ob ein Alarmzeitpunkt erreicht ist
#ifdef LCD
		DCF_show_alarm();						// zeigen, wieviele Alarmdefinitionen definiert/aktiv ist
#endif
		error = 0;								// Fehlerflags der letzten Minute zurcksetzen
													// aber erst nach der Aktualisierung des Displays
		}
	}												// while(1)
}


// -----------------------------------------------------------------------------
// INT1 - rising edge = Ende eines Impulses
// Es ist die Dauer des Impulses zu bestimmen (0.1 oder 0.2 Sekunden)
// schreibt das empfangene Bit direkt in das Bitarray.
// bei Fehler werden Flags in error gesetzt, so dass man nach Ablauf einer
// vollstndigen Minute sehen kann, welche Fehler aufgetreten sind,
// allerdings nicht, wie oft !
ISR(INT1_vect)
{
register uint8_t x;
x = T1_ticks;
/*
_delay_us(100);								// kurz warten und testen, ob der PIN wirklich
if (!(PIND & (1<<PD3))) return;			// high ist oder ob es eine Strung war													
*/
													// krzer als 0.1 Sekunden
if      (x <  (MS_100 - TOL1)) error |= (1<<ERR_DCF_01);
													// 0.1 Sekunden
else if (x <  (MS_100 + TOL1)) DCF_save_bit(0); // eine 0 speichern
													// krzer als 0.2 Sekunden
else if (x <  (MS_200 - TOL1)) error |= (1<<ERR_DCF_02);
													// 0.2 Sekunden
else if (x <  (MS_200 + TOL1)) DCF_save_bit(1); // eine 1 speichern
													// lnger als 0.2 Sekunden
else if (x >= (MS_200 + TOL1)) error |= (1<<ERR_DCF_03);
}

// -----------------------------------------------------------------------------
// INT0 - falling edge = Start des Impulses
// Es wird geprft, wieviel Zeit seit dem letzten Impulsbeginn vergangen ist,
// es knnen 1 Sekunde (normales DCF_Bit) oder 2 Sekunden (Minutenbeginn) sein.
// Wesentliche Aufgabe des Interrupts ist, den 2-Sekunden-Signalabstand zu Beginn
// einer neuen Minute zu erkennen und darauf die Quarz-Uhr zu synchronisieren
// und den Beginn einer neuen Minute an main() zu signalisieren.
// Sind Empfangsfehler im DCF-Signal aufgetreten, dann unterbleibt diese Synchronisation.
// Es lst in diesem Fall (error > 0) der 1-Sekundentimer0 den Minutenbeginn aus.
ISR(INT0_vect)
{
register uint8_t x;
x = T1_ticks;									// volatile laden
/*
_delay_us(100);								// kurz warten und testen, ob der PIN wirklich
if (PIND & (1<<PD2)) return;				// low ist oder ob es eine Strung war
*/
													// krzer als 1 Sekunde
if      (x <  (MS_1000 - TOL2)) error |= 1<<ERR_DCF_10;
													// 1 Sekunde - nichts tun, das ist der Normalfall
else if (x <  (MS_1000 + TOL2));
													// krzer als 2 Sekunden
else if (x <  (MS_2000 - TOL2)) error |= 1<<ERR_DCF_20;
											
else if (x <  (MS_2000 + TOL2))			// 2 Sekunden
	{
	if (!(error))
		{											// Wenn kein Fehler vorlag, interne Uhr am Minutenbeginn		
		TCNT0 = 0;								// synchronisieren 
		T0_ticks = T0_PRELOAD;				// Timer0 decrementiert T0_ticks auf 0 fr 1 Sekunde
		flag  |= (1<<DCF_MINUTE );			// Flag fr neue DCF-Minute	
		}
	}	

else error = ERR_NO_DCF;					// lnger als 2 Sekunden und damit ungltig

T1_ticks = 0;									// Counter fr Timer1 zurcksetzen
TCNT1H   = 0;
TCNT1L   = 0;
}

// -----------------------------------------------------------------------------
// wird als interne Zeitreferenz fr den Sekundentakt genutzt
// Prescaler = 64, Preload = 256, Counter 225, 3686800 / (64 * 256 * 225) = 1 Sekunde
// Das Flag 1<<CLK_SEC  wird nach genau 1 Sekunde gesetzt und wird in main() ausgewertet.
// Bei fehlerfreiem DCF-Signal signalisiert das 2-Sek-DCF-Signal den Minutenstart.
// Bei fehlerhaftem DCF-Signal signalisiert die 60. Sekunde des Timer0 den Minutenstart.

ISR(TIMER0_OVF_vect)
{

if (--T0_ticks) return;						// abbrechen solange T0_ticks nicht zu 0 decrementiert wurde				
T0_ticks = T0_PRELOAD;						// neuen Startwert setzen

time[CLK][SEKUNDE]++;						// Sekunde incrementieren

if (time[CLK][SEKUNDE] < 59)				// solange < 59 wird nur die Sekunde angezeigt
	{												// hier gibt der Timer den Takt an
	flag |= (1<<CLK_SEC);					// Flag fr neue Sekunde setzen
	}

else if (time[CLK][SEKUNDE] == 59) 		// in der 59. Sekunde werden zwei Flags gesetzt:
	{												// - CLK_SEC zeigt die neue Sekunde an,
	flag |= (1<<CLK_SEC_59 | 1<<CLK_SEC);//- CLK_SEC_59 wertet das vollstndige DCF-Signal aus
	}

else if (error)								// falls im timing des DCF-Signals Fehler aufgetreten sind
	{												// dann lutet der Timer0 die neue Minute ein
	flag |= (1<<CLK_MINUTE);				// Flag fr neue Minute durch CLK setzen
	}
}

// -----------------------------------------------------------------------------
// fr die Bestimmung der Impulslngen, wird alle 0,01 Sekunden ausgelst
// bei normaler Funktion werden 100/200 ticks ausgelst
// kommt kein Signal, dann luft ticks ber, in diesem Fall wird der Wert
// konstant auf 255 gehalten.
// An diesem Wert kann erkannt werden, dass ein DCF-Signal berfllig ist.
ISR(TIMER1_COMPA_vect)
{
TCNT1H = 0;
TCNT1L = 0;

T1_ticks++;										// incrementieren bis zum berlauf auf 0
if (!(T1_ticks))
	{
	T1_ticks--;									// wenn T1_ticks berluft, dann zurcksetzen auf 255
	error = T1_ticks;							// Hier wird der Fehlercode 0xFF = ERR_NO_DCF gesetzt:
	}												// fehlendes DCF-Signal
}

// ----------------------------------------------------------------------------
// Dekodiert die gelesenen 58 Bits und schreibt die Daten der nchsten Minute
// ins SRAM in das Array 'DCF_time[]'.
// Die zugehrigen Parittsbits werden berprft, Fehler werden in Error.0/1/2 markiert.
// Die Ankndigungs- und Zusatzbits werden in DCF_time[ZBITS] abgelegt.
// Die Ankndigungsbit stehen im Signal ab Pos. 15 in der Reihenfolge:
// R, A1, Z1, Z2, A2, S  (wobei S immer = 1).
// Sie werden umsortiert (warum eigentlich ?):
// DCF_zbits:(..|..|..|A1|A2|R |Z1|Z2)
// Sommerzeit(..|..|..|A1|A2|R | 1| 0)
// Winterzeit(..|..|..|A1|A2|R | 0| 1)
// A1 Ankndigungsbit fr Sommerzeit
// A2 Ankndigungsbit fr Schaltsekunde
// R  Reserveantenne aktiv
// Die Sommerzeit wird direkt als Bit.7 in time[CLK][DOW] abgelegt.

void DCF_decode(void)
{
													// zuerst die Zusatzbits nach DCF_time[ZBITS] schreiben
time[DCF][ZBITS]  = DCF_get_bit(16);			// A1
time[DCF][ZBITS] *= 2;						// nach links schieben, Platz fr nchstes Bit machen
time[DCF][ZBITS] += DCF_get_bit(19);			// A2
time[DCF][ZBITS] *= 2;						// nach links schieben, Platz fr nchstes Bit machen
time[DCF][ZBITS] += DCF_get_bit(15);			// R
time[DCF][ZBITS] *= 2;						// nach links schieben, Platz fr nchstes Bit machen
time[DCF][ZBITS] += DCF_get_bit(17);			// Z1
time[DCF][ZBITS] *= 2;						// nach links schieben, Platz fr nchstes Bit machen
time[DCF][ZBITS] += DCF_get_bit(18);			// Z2

time[DCF][MINUTE] = calc_time(21, 21 + 7);// Minute auslesen, Paritt prfen
if (calc_parity(21, 21 + 8)) error |= (1<<ERR_P_MINUTE);

time[DCF][STUNDE] = calc_time(29, 29 + 6);// Stunde auslesen, Paritt prfen
if (calc_parity(29, 29 + 7)) error |= (1<<ERR_P_HOUR);

time[DCF][TAG] = calc_time(36, 36 + 6);// Tag auslesen

time[DCF][DOW] = calc_time(42, 42 + 3);// Day of week auslesen
													// die Nr. des Tag steht nun als Ziffer in DOW
													// z.B. Mittwoch = 3
													// umwandeln in Bitposition 0 ... 6
													// dazu die 1 um (3-1) mal nach links shiften
time[DCF][DOW] = 1 << (time[DCF][DOW] - 1);

													// die Sommerzeit als Bit.7 in den DOW schmuggeln
if ((time[DCF][ZBITS] & 0x03) == 2) set_summertime(&time[DCF][0], TRUE);

time[DCF][MONAT] = calc_time(45, 45 + 5);// Monat auslesen

time[DCF][JAHR] = calc_time(50, 50 + 8);// Jahr auslesen, Paritt prfen
if (calc_parity(36, 36 + 23)) error |= (1<<ERR_P_REST);
}

// -----------------------------------------------------------------------------
// Decodiert eine Zahl aus einem Bit-Bereich des DCF_Bit_Arrays,
// das ergibt dann den Wert fr entweder: Minute, Stunde, Tag, Monat, Jahr
// Parameter: Startposition des ersten Bits
//		        Endposition   des letzten Bits
// Rckgabe:  der ermittelte Wert
uint8_t calc_time(uint8_t start, uint8_t end)
{
// uint8_t summand[] PROGMEM = { 1, 2, 4, 8, 10, 20, 40, 80 };
// um PROGMEM zu nutzen, darf das Array nicht lokal sein

uint8_t r = 0;
uint8_t i = 0;

while (start < end)
	{
	if (DCF_get_bit(start++))
		{
		r += pgm_read_byte(&summand[i]);
		}
	i++;
	}

return(r);
}

// -----------------------------------------------------------------------------
// Berechnet die Paritt, beginnend an Position "start" ber "cnt" Bits
// (Achtung:  das letzte Bit im Bereich ist das Parittsbit !)
// Parameter: Startpunkt ab Arraybeginn
//		        Letzes Bit = Parittsbit
// Rckgabe:  (0 = kein Parittsfehler, 1 = Parittsfehler)
// Das Parittsbit ergnzt die Summe der einzelnen Bits auf eine gerade Zahl,
// d.h. Bit.0 ist nicht gesetzt - sonst liegt ein Fehler vor.
uint8_t calc_parity(uint8_t start, uint8_t end)
{
register uint8_t p = 0;
while (start < end)
	{
	if (DCF_get_bit(start++)) p++;
	}

return(p & 0x01);
}

// -----------------------------------------------------------------------------
// Ein Zeitarray wird um x Minuten incrementiert.
// ES wird die Zeitzhlung aktualisiert und incl. aller mglichen berlufe.
// Weiterhin erfolgt Prfung auf Wechsel zwischen Sommer-/Winterzeit.
void DCF_incr_time(uint8_t *zeit, uint8_t x)
{
register uint8_t t;

do	{
	zeit[MINUTE]++;
	if (zeit[MINUTE] < 60) continue;		// solange keine volle Stunde angebrochen ist,
													// kann hier beendet werden
	zeit[MINUTE] = 0;
	zeit[STUNDE]++;
													// zu jeder vollen Stunde testen, ob eine Wechsel
	switch_summertime(zeit);				// zwischen Sommer- und Winterzeit ansteht.
													// im jeweiligen Zeitarray wird dann das Flag fr
													// die Sommerzeit (DOW.7) gegebenfalls gendert
													// wenn Stunde == 24, dann kann kein Fehler entstehen,
													// da nur um 2 bzw. 3 Uhr umgeschaltet wird
	if (zeit[STUNDE] < 24) continue;
	zeit[STUNDE] = 0;
	zeit[TAG]++;
													// den Wochentag incrementieren							
	t = zeit[DOW] & 0x80;					// Bit.7 im Wochentag markiert die Sommerzeit, zwischenspeichern !
	zeit[DOW] &= 0x7F;						// Bit.7 entfernen
	zeit[DOW] *=2;								// Wochentag incrementieren (nach links schieben)
													// berlauf von Sonntag auf Montag
	if (zeit[DOW] == 0x80) zeit[DOW] = 0x01;
	zeit[DOW] += t;							// Sommerzeit als Bit.7 wieder addieren

													// nun kommen die Tage des Monats in Spiel ...
	if (zeit[MONAT] == 2)					// zuerst Sonderfall Februar und Schaltjahr
		{
		if (is_leap(zeit[JAHR]))
			{
			if (zeit[TAG] <= 29) continue;
			}
		else
			{
			if (zeit[TAG] <= 28) continue;
			}
		}

	else											// die brigen Monate
		{
		if (zeit[TAG] <= pgm_read_byte(&days_of_month[zeit[MONAT]])) continue;
		}

	zeit[TAG] = 1;								// der letzte Tag des Vormonats war erreicht
	zeit[MONAT]++;
	if (zeit[MONAT] > 12)
		{
		zeit[MONAT] = 1;
		zeit[JAHR]++;
		}
	} while(--x);								// den Rundgang fr jede Minute wiederholen
}

// ----------------------------------------------------------------------------
// Die fr die Bestimmung des Sonnenstandes bentigte UTC-Zeit liegt in unserer
// Zeitzone im Winter um 1 Stunde, whrend der Sommerzeit um 2 Stunden vor unserer Zeit
// Kopiert die CLK_time auf die UTC_time und zhlt je nach cet/cest um eine bzw. zwei
// Stunden zurck.
void DCF_make_UTC_time(void)
{
DCF_copy_time(&time[UTC][0], &time[CLK][0]);	// die Zeit kopieren

DCF_decr_time(&time[UTC][0], 60);				// um 60 Minuten decrementieren
if (is_summertime(&time[CLK][0]))				// und whrend der Sommerzeit um eine weitere Stunde
	{
	DCF_decr_time(&time[UTC][0], 60);			// um noch einmal 60 Minuten decrementieren
	}
}

// ----------------------------------------------------------------------------
// decrementiert ein Zeitarray um x Minuten
// *zeit ist das Zeitarray, das bearbeitet werden soll
void DCF_decr_time(uint8_t *zeit, uint8_t x)
{
register uint8_t t;

do	{
	if (zeit[MINUTE]) 							// solange die Minute > 0 gibt es keinen berlauf
		{
		zeit[MINUTE]--;
		continue;									// das war's schon
		}
	else
		{
		zeit[MINUTE] = 59;
		zeit[STUNDE]--;
		}

	if (zeit[STUNDE] == 255) 					// auf berlauf testen
		{
		zeit[STUNDE] = 23;
		}
	else
		{
		continue;
		}

	// den Wochentag decrementieren							
	t = zeit[DOW] & 0x80;						// das Bit fr die Sommerzeit sichern
	zeit[DOW] &= 0x4F;							// die Sommerzeit entfernen
	zeit[DOW] /= 2;								// den DOW decrementieren
														// auf underflow testen
	if (!(zeit[DOW])) zeit[DOW] = 0x40; 	// wenn == 0, dann DOW.6 setzen
	zeit[DOW] += t;								// Sommerzeit wieder addieren

	// den Test auf eine nderung der Sommerzeit erspare ich mir an dieser Stelle ...
	// ist fr den Sonnenstand nicht von Bedeutung, da die UTC-Zeit massgeblich ist
	if (zeit[TAG] > 1)
		{
		zeit[TAG]--;
		continue;
		}

	else												// bei Underflow des Tages zuerst den
		{												// neuen Monat bestimmen
		zeit[MONAT]--;
		if (zeit[MONAT] == 0)
			{
			zeit[MONAT] = 12;
			zeit[JAHR]--;
			}											// nun haben wir die Anzahl der Tage
		zeit[TAG] = pgm_read_byte(days_of_month[zeit[MONAT]]);
														// falls aber Februar und Schaltjahr
		if ((zeit[MONAT] == 2) && (is_leap(zeit[JAHR]))) zeit[TAG]++;
		}
	} while (--x);
}

// -----------------------------------------------------------------------------
// kopiert ein Zeit-Array
void DCF_copy_time(uint8_t *target, uint8_t *source)
{
register uint8_t i = 6;							// Element mit Index 5 ... 0 kopieren
do	{
	i--;
	target[i] = source[i];
	} while(i);
}

// -----------------------------------------------------------------------------
// vergleicht die Bytes der Arrays time1 und time2 byteweise miteinander
// liefert: TRUE  wenn identisch,
//          False wenn ungleich
uint8_t DCF_compare_time(uint8_t *time1, uint8_t *time2)
{
uint8_t i = 6;										// Element mit Index 5 ... 0 vergleichen
uint8_t r = TRUE;
do {
	i--;												// ein einziger fehlerhafter Vergleich reicht aus
	if (! (time1[i] == time2[i]) ) r = FALSE;
	} while (i);
return(r);
}

// -----------------------------------------------------------------------------
// Wird sinnvollerweise jeweils zu einer vollen Stunde aufgerufen, da nur dann ein Wechsel
// zwischen Sommer- und Winterzeit stattfinden kann.
//
// Prfen anhand von Datum und Zeit der internen Uhr, ob zwischen Sommer- und Winterzeit umgeschaltet werden muss.
//
// Am letzten Sonntag im Mrz um 02.00h wird die Uhr auf 03.00h,
// am letzten Sonntag im Oktober um 03.00h wird die Uhr auf 02.00h gestellt.
// Ausgehend von der aktuellen Sommerzeit wird geprft,
// wenn Sommerzeit = 0, ob Monat  = 3, Wochentag = 7 und Datum zwischen 25 und 31 und Stunde = 2
// wenn Sommerzeit = 1, ob Monat = 10, Wochentag = 7 und Datum zwischen 25 und 31 und Stunde = 3
// zur Kontrolle/Besttigung knnte noch das Ankndigungsbit A1 (Bit 16) herangezogen werden ...
// kritisch ist Umschaltung im Oktober, da hier der Zeitraum von 0.200h bis 03.00h zweimal nacheinander
// durchlaufen wird -> hier die aktuelle Einstellung der Sommerzeit mageblich:
// wenn Sommerzeit, dann Umschaltung, wenn (schon) Winterzeit, dann keine Umschaltung !
void switch_summertime(uint8_t *zeit)
{
if (zeit[TAG] > 24)							// Gewechselt wird zwischen dem 25. und dem 31. (Mrz bzw. Oktober)
	{
	if ((zeit[DOW] & 0x40) )				// nur an einem Sonntag (dann ist Bit.6 gesetzt)
		{
		if (zeit[MONAT] == 3)				// im Mrz
			{
			if (zeit[STUNDE] == 2)			// um 2:00h die Uhr auf 3:00h vorstellen
				{
				zeit[STUNDE] = 3;				// Zeit vorstellen auf 3;00h
				set_summertime(zeit, TRUE);// auf Sommerzeit schalten
				}
			}
													// im Oktober die Uhr zurckstellen (sofern noch Sommerzeit !)
													// 03.00h wird zweimal durchlaufen,
													// beim 2. Mal ist aber bereits auf Winterzeit umgeschaltet,
													// daher wird die Uhr nicht noch einmal zurckgestellt
		else if ((zeit[MONAT] == 10) && (is_summertime(zeit)))
			{
			if (zeit[STUNDE] == 3)
				{
				zeit[STUNDE] = 2;				// Zeit zurckstellen auf 2:00h
				set_summertime(zeit, FALSE);// und auf Winterzeit stellen
				}
			}
			
		}
	}
}

// ----------------------------------------------------------------------------
// Prft, ob das DCF-Signal Sommerzeit signalisiert
// Rckgabe: 1 = Sommerzeit
//           0 = Keine Sommerzeit
// es wird nicht DCF_zbits ausgewertet, sondern das Bit.7 in time[DOW] !
uint8_t is_summertime(uint8_t *zeit)
{
return(zeit[DOW] & 0x80); 					// Bit.7 in DOW signalisiert die Sommerzeit
}

// -----------------------------------------------------------------------------
// setzt/lscht das Sommerzeitflag in time[DOW].7
void set_summertime(uint8_t *zeit, uint8_t i)
{
if (i) zeit[DOW] |= 0x80; else zeit[DOW] &= ~(0x80);
}

// -----------------------------------------------------------------------------
// ein Schaltjahr liegt vor, wenn die Jahreszahl durch 4, 100 (oder 400) ohne Rest teilbar
// es werden nur die beiden Flle mod 4/ mod 100 untersucht

uint8_t is_leap(uint8_t j)
{													// Es liegt ein Schaltjahr vor,
if (!(j & 0x03)) return(TRUE);			// wenn mod 4 keinen Rest ergibt

while (j > 100) j -= 100;					// diese berprfung ist hier wohl eher berflssig ..
if (!(j)) return(TRUE);						// wenn mod 100 keinen Rest ergibt

return(FALSE);
}


// -----------------------------------------------------------------------------
// Ermittelt die lfd. Nummer des aktuellen Tages im Jahr unter Bercksichtigung von Schaltjahren
uint16_t day_of_year(uint8_t *zeit)
{
uint16_t z = 0;
uint8_t i = 1;

while (i < zeit[MONAT])						// die vollen Tage der Monate addieren
	{
	z+= pgm_read_byte(&days_of_month[i++]);
	if (i == 2)									// im Februar das Schaltjahr bercksichtigen
		{
		if (is_leap(zeit[JAHR])) z++;		// und einen Tag addieren fr den 29. Februar
		}
	}
z += zeit[TAG];								// nun noch die Tage des laufenden Monats addieren
return(z);
} 

// -----------------------------------------------------------------------------
// bei Anbruch einer neuen Sekunde nur die Sekundenanzeige aktualisieren
#ifdef LCD
void display_second(void)
{
lcd_goto_xy(1,19);
lcd_dezprint(time[CLK][SEKUNDE]);		// Sekunde dezimal ausgeben

//DCF_show_alarm();							// sekndlich Vernderung der Schaltdefinition ausgeben

}
#endif

// -----------------------------------------------------------------------------
// die gesamte Zeitanzeige Stunden, Minuten (und Sekunden) aktualisieren
#ifdef LCD
void display_time(uint8_t *zeit)
{

register uint8_t i = 0;
register uint8_t j;

lcd_goto_xy(1, 1);

j = (zeit[DOW] & 0x7F);						// Bit.7 ignorieren -> Flag fr Sommerzeit
													// der Tag ist als Bit 1 ..6 codiert !
			
while (j)										// Index des Textstrings bestimmen
	{												// dazu die binrzahl solange nach rechts
	j = j >> 1;				
	if (j) i++;
	}
i = i * 2;										// es werden 2 Char je Tag benutzt
	
lcd_putc(pgm_read_byte(&day_names[i++]));
lcd_putc(pgm_read_byte(&day_names[i]));
lcd_putc(' ');

lcd_dezprint(zeit[TAG]);
lcd_putc('.');
lcd_dezprint(zeit[MONAT]);
lcd_putc('.');
lcd_dezprint(zeit[JAHR]);

lcd_putc(' ');

lcd_dezprint(zeit[STUNDE]);
lcd_putc(':');
lcd_dezprint(zeit[MINUTE]);
lcd_putc(':');
lcd_dezprint(time[CLK][SEKUNDE]);

lcd_goto_xy(2,6);							// den Unterstrich in Zeile 2 ziehen
i = 15;										// nicht in dem Bereich, in dem die Alarmdefs
while (i--)									// angezeigt werden (Spalte 1-5)
	{
	lcd_putc('_');
	}

lcd_goto_xy(2,13);						// Position fr Fehler/Sommmer/Winterzeit
if (error) 									// wenn in der letzten Minute ein Fehler auftrag
	{
	lcd_puts_p(PSTR("er:"));			// Fehlermeldung
	lcd_hexprint(error);					// Fehlercode hexadezimal
	}

else											// ohne Fehler wird Hinweis auf Sommerzeit ausgegeben
	{
	if (is_summertime(&time[CLK][0])) lcd_puts_p(PSTR("cest")); else lcd_puts_p(PSTR("cet"));
	}

#ifdef SHOW_ERRORS
char space[] = "  ";
lcd_goto_xy(3, 1);
if (time[DCF][ZBITS] & ~(0x03))		// wenn Ankndigungsbits gesendet wurden
												// die Bits fr Sommer-/Winterzeit ignorieren 
	{
	if (!(error))							// nur Anzeigen, wenn das Datum fehlerfrei
		{
		lcd_hexprint(time[DCF][ZBITS]);
		}
	}

else
	{
	lcd_puts(space);						// den Bereich der Meldung lschen	
	}

if (error_abs)								// absolute Anzahl der Fehler ausgeben
	{
	lcd_goto_xy(4, 1);

	if ((time[CLK][MINUTE] == 0) && (time[CLK][STUNDE] == 0)) // Mitternacht
		{
		lcd_puts(space);					// um Mitternacht den Zhler und die
		error_abs = 0;						// Anzeige der Fehler lschen

		lcd_goto_xy(3, 1);				// auch die Anzeige der Ankndigungsbits lschen
		lcd_puts(space);			
		}

	else
		{
		lcd_hexprint(error_abs);		// Anzahl der kumulierten Fehler des Tages ausgeben
		}
	}
#endif

}
#endif



							

// --- [ eof ] ----------------------------------------------------------------
