// --- [ dcf_77.h ] -----------------------------------------------------------
//       01.05.09
//       tab = 3
//
// ----------------------------------------------------------------------------

// --- Konstanten -------------------------------------------------------------

#define ALARMPORT		PORTD					// frei: PORTD.0.1.4.5.6.7
#define DAYLIGHT_PIN PD7					// benutzt wird zur Zeit nur PIN 7 !

#define MS_100			9						// Counts fr 0.1 Sek Puls [Einheit 0.01sec]
#define MS_200			19						// Counts fr 0.2 Sek Puls
#define MS_1000		99						// Counts fr 1.0 Sek Puls
#define MS_2000		199					// Counts fr 2.0 Sek Puls
#define TOL1			3						// zulssige Zeitabweichung +- des DCF-Signals (bei 0.1/0.2 Sekunden)
#define TOL2			5						// zulssige Zeitabweichung +- des DCF-Signals (bei 1/2 Sekunden)

													// Bitpositionen fr flag
#define CLK_SEC		1						// neue Sekunde ist angebrochen
#define CLK_SEC_59	2						// 59. Sekunde der internen Uhr erreicht
#define CLK_MINUTE	3						// neue Minute der internen Uhr erreicht (bei fehlendem DCF-Signal)
#define DCF_MINUTE	4						// neue Minute aus DCF-Signal
#define DISPLAY		5						// die Anzeige muss aktualisiert werden
#define SYNCH			6
#define START_FLAG	7						// Signalisiert den Programmstart, bentigt zur bernahme der 1. Minute

#define ERR_P_MINUTE 0						// Parittsfehler bei Minuten
#define ERR_P_HOUR	1						// Parittsfehler bei Stunden
#define ERR_P_REST	2						// Parittsfehler bei restlichen Daten
#define ERR_DCF_01	3						// DCF_Fehler - Impuls krzer als 0.1 Sekunde
#define ERR_DCF_02	4						// DCF_Fehler - Impuls krzer als 0.2 Sekunde
#define ERR_DCF_03	5						// DCF_Fehler - Impuls lnger als 0.2 Sekunde
#define ERR_DCF_10	6						// DCF_Fehler - Impulsabstand lnger/krzer als 1 Sekunde
#define ERR_DCF_20	7						// DCF_Fehler - Impulsabstand lnger als 2 Sekunde
#define ERR_COMP_DCF	0xFE					// Fehler beim Zeitvergleich zweier aufeinanderfolgender Minuten
#define ERR_NO_DCF	0xFF					// Signalabstand grer als 2 Sekunden
#define ERR_SYNCH		0xFD					// Durch Zeitvergleich mit letzter DCF-Minute wird neue Zeit geschrieben
#define ERR_TWI   	0xFC					// Die Uhrzeit wurde via TWI gendert



// --- Prototypen -------------------------------------------------------------

void SUN_display(void);
void TWI_SLA_Initialise(void);

void DCF_decode(void);
uint8_t calc_time(uint8_t i, uint8_t cnt);
uint8_t calc_parity(uint8_t i, uint8_t cnt);
void DCF_decr_time(uint8_t *time, uint8_t minute);
void DCF_incr_time(uint8_t *time, uint8_t minute);
void DCF_copy_time(uint8_t *target, uint8_t *source);
void DCF_make_UTC_time(void);
uint8_t DCF_compare_time(uint8_t *time1, uint8_t *time2);
void DCF_check_alarm(void);
void DCF_check_daylight(void);
void DCF_set_alarm(uint8_t status);
void DCF_show_alarm(void);
void DCF_read_alarm(void);
void DCF_delete_alarm(void);
void DCF_delete_eeprom(void);
void DCF_write_alarm(void);
uint8_t is_leap(uint8_t j);
uint16_t day_of_year(uint8_t *time);
void switch_summertime(uint8_t *time);
uint8_t is_summertime(uint8_t *time);
void set_summertime(uint8_t *time, uint8_t i);
void display_second(void);
void display_time(uint8_t *time);
void zeit_zeichen(void);

#ifdef LOG_RS232
void print_datum(void);
void print_dcfbits(void);
void print_alarm(uint8_t idx);
#endif

// --- [ eof ] ----------------------------------------------------------------